/*
 * ctcompare: Compare two C tokens files for similarities.
 * Copyright (c) Warren Toomey, under the GPL license.
 * 
 * $Revision: 1.16 $
 */
#include <stdio.h>
#include <sys/stat.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include "tokens.h"
#define ISOMORPH_CODE		/* Turn on to enable isomorphic code compares */
#define OPT_OVERLAP		/* Turn on to enable overlap optimisations */
#undef DEBUG

#define STARTFILE1 0xfe		/* Token id for start of each srcfile */
#define STARTFILE2 0xff		/* Token id for start of each dstfile */
#define BUF_SIZE 1024

/*
 * Each token is represented by the following structure. Each file in a ctf
 * file is represented as a linked list following the next pointer. The value
 * holds the token's value for two token types: identifiers and numbers.
 */
struct token {
  char *file;			/* File this token comes from */
  unsigned int linenum;		/* Line # on which token appeared */
  struct token *next;
  unsigned short value;
  unsigned char type;
};

/* When we find a token run in each source file, we keep a linked list
 * of the run in this node.
 */
struct tokrun {
  int sstart, send;
  int dstart, dend;
  struct tokrun *next;
};

/* Token intarrays */
int *Toksrcarray;
int *Tokdstarray;
int Toksrcsize;
int Tokdstsize;
struct token **Toksrcref;
struct token **Tokdstref;

/* Run-time parameters */
int threshold = 20;		/* Threshold above which we report a run */
int showtokens = 0;		/* If 1, print out tokens in a run */
int crosscompare = 0;		/* If 1, print out actual code in a run */
int progress = 0;		/* Print filenames to stderr as we process */
int skip_cppstuff = 1;		/* Skip #defines, #includes etc. */

/* Array used to print out a run of tokens */
unsigned char *tokstring[] = {
  "ERR ", "ERR ", "ERR ", "ERR ", "ERR ", "ERR ", "ERR ",
  "ERR ", "ERR ", "ERR ", "\n", "ERR ", "ERR ", ">>= ", "ERR ", "ERR ",
  "ERR ", "ERR ", "ERR ", "ERR ", "ERR ", "ERR ", "ERR ", "ERR ", "ERR ",
  "ERR ", "ERR ", "ERR ", "ERR ", "ERR ", "ERR ", "ERR ", "/= ", "! ",
  "\"string", "->", "++ ", "% ", "& ", "'c", "( ", ") ", "* ", "+ ", ", ", "- ",
  ". ", "/ ", "-- ", "&& ", "|| ", "++= ", "%= ", "-= ", "&= ", "*= ", "|= ",
  "NUM", ": ", "; ", "< ", "= ", "> ", "? ", "!= ", "<= ", "case ", "char ",
  "const ", "continue ", "default ", "do ", "...", "double ", "else ",
  "enum ", "extern ", "float ", "for ", "goto ", "if ", "int ", "long ",
  "register ", "return", "short ", "signed ", "sizeof ", "static ", "struct ",
  "switch ", "[ ", "\\", "] ", "^ ", "id", "label", "typedef ", "union ",
  "unsigned ", "void ", "volatile ", "while ", "#define ", "#elif ",
  "#else ", "#endif ", "#error ", "#ifdef ", "#if ", "#ifndef ", "#include ",
  "#line ", "#pragma ", "#undef ", "#warning ", "~= ", "== ", "break ", ">= ",
  "<< ", ">> ", "<<= ", "{ ", "| ", "} ", "~ ", "ERR "};


int
build_intarray(struct token * head, int issecond, int numtokens)
{
  unsigned int i, group;
  struct token *T;
  struct token *oldT;		/* This is the token 3 back */

  numtokens+=8;			/* See the bit about seeding below! */

  /* Allocate the array to holds the integerised tokens */
  if (issecond==1) {
    Tokdstsize = numtokens;
    Tokdstarray = (int *)malloc(numtokens * sizeof(int));
    if (Tokdstarray == NULL) return (0);
    Tokdstref= (struct token **)malloc(numtokens * sizeof(struct token *));
    if (Tokdstref == NULL) return (0);
  } else {
    Toksrcsize = numtokens;
    Toksrcarray = (int *)malloc(numtokens * sizeof(int));
    if (Toksrcarray == NULL) return (0);
    Toksrcref= (struct token **)malloc(numtokens * sizeof(struct token *));
    if (Toksrcref == NULL) return (0);
  }


  /* The most significant octet in each group is the token */
  /* at that index in the Tokarray, i.e the top octet in */
  /* Toksrcarray[100] represents source token 100 */

  /* We build the first group of 4 tokens manually */
  /* Danger: there had better be 4 tokens in the list :-) */
  oldT = T = head;
  group = T->type; T = T->next;
  group = (group << 8) | T->type; T = T->next;
  group = (group << 8) | T->type; T = T->next;
  group = (group << 8) | T->type;
  if (issecond==1) {
    Tokdstarray[0] = group;
    Tokdstref[0] = oldT;
  } else {
    Toksrcarray[0] = group;
    Toksrcref[0] = oldT;
  }

  T = T->next; oldT = oldT->next;

  /* Now walk the list and build more groups */
  for (i = 1; i < numtokens; i++) {
    if (T != NULL) {
      group = (group << 8) | T->type;
      if (issecond==1) {
      	Tokdstarray[i] = group;
        Tokdstref[i] = oldT;
      } else {
      	Toksrcarray[i] = group;
        Toksrcref[i] = oldT;
      }
      T = T->next; oldT = oldT->next;
    } else {

      /* We seed the end with patterns that are */
      /* guaranteed to be different for both lists. */
      /* This saves us having to keep track of lengths later. */
      group = (group << 8) | issecond;
      if (issecond==1) {
      	Tokdstarray[i] = group;
        Tokdstref[i] = oldT;
      } else {
      	Toksrcarray[i] = group;
        Toksrcref[i] = oldT;
      }
      if (oldT->next!=NULL) oldT = oldT->next;
    }
  }
  return (1);
}


/*
 * Load the ctf file, building token nodes and putting them into a linked
 * list using the next pointer. The end of each file is represented by an
 * ENDFILE node. Returns a pointer to a linked list of nodes, or NULL if
 * something goes wrong.
 */
struct token *
load_ctffile(char *filename, int issecond)
{
  FILE *zin;
  char buf[BUF_SIZE];
  unsigned char type;
  int i, ch, linenumber = 0;
  int numtokens = 0;
  unsigned short value;
  char *fname = NULL;
  int inside_cpp_line = 0;

  struct token *this = NULL, *last = NULL, *head = NULL;

  /* Open the file */
  zin = fopen(filename, "r");
  if (zin == NULL) {
    fprintf(stderr, "Can't open %s\n", filename); return (NULL);
  }

  /* Check the ctf header */
  if ((fgetc(zin) != 'c') || (fgetc(zin) != 't') || (fgetc(zin) != 'f') ||
      (fgetc(zin) != 0x1) || (fgetc(zin) != 0x2)) {
    fprintf(stderr, "File %s is not a ctf file\n", filename);
    exit(1);
  }

  /* Now loop reading in the file */
  while ((ch = fgetc(zin)) != EOF) {

    switch (ch) {
    case IDENTIFIER:		/* We need to extract the */
    case INTVAL:		/* values of these 2 types */
      value = (fgetc(zin) & 0xff) << 8;
      value += (fgetc(zin) & 0xff);
      break;
    default:
      value = 0;
    }

    switch (ch) {
    case LINE:
      linenumber++; inside_cpp_line = 0;
      break;


    case HASHdefine:		/* Ignore C preprocessor tokens */
    case HASHelif:		/* if skip_cppstuff==1 */
    case HASHelse:
    case HASHendif:
    case HASHerror:
    case HASHifdef:
    case HASHif:
    case HASHifndef:
    case HASHinclude:
    case HASHline:
    case HASHpragma:
    case HASHundef:
    case HASHwarning:
      inside_cpp_line = 1;

    default:
      if ((skip_cppstuff == 1) && (inside_cpp_line == 1)) break;

#ifdef DEBUG
      printf("Got token %s\n", tokstring[ch]);
#endif

      /* Build a token node */
      last = this;
      if (head == NULL) head = this;
      this = (struct token *) malloc(sizeof(struct token));
      if (this == NULL) return (NULL);

      if (last != NULL) last->next = this;
      this->next = NULL;

      /* If this is a FILE token, read the filename in */
      if (ch==FILENAME) {
        for (i = 0; i < BUF_SIZE - 1; i++) {
	  if (((ch = fgetc(zin)) == EOF) || (ch == '\0')) break;
	  buf[i] = ch;
        }
        buf[i] = '\0';
	fname= strdup(buf);
#ifdef DEBUG
        printf("Just found filename %s\n", fname);
#endif
        if (issecond) this->type = STARTFILE2;
        else this->type = STARTFILE1;
        linenumber = 1;
	value=0;
      } else {
        this->type = type = (unsigned char)ch;
      }

      this->value = value;
      this->file = fname;
      this->linenum = linenumber;
      last = this;
      numtokens++;
    }
  }

  /* The file's finished, close it */
  fclose(zin);

  /* Build an int array for quick comparisons */
  if (build_intarray(head, issecond, numtokens) == 0) return (NULL);
  return (head);
}


/*
 * Print out lines of two files side-by-side. This is best viewed with a
 * terminal window 160 characters wide!
 */
void
paste_files(char *file1, int start1, int num1,
	    char *file2, int start2, int num2)
{
  FILE *f1in, *f2in;
  char buf[1024];
  int count_to_eighty=0;
  int i, bptr, maxlines;
  int tab_upto=0;

  f1in = fopen(file1, "r");
  if (f1in == NULL) return;
  f2in = fopen(file2, "r");
  if (f2in == NULL) { fclose(f1in); return; }

  /* Get maximum number of lines */
  maxlines = (num1 > num2) ? num1 : num2;
  start1--; start2--;

  /* Read up to line start1 in file1 */
  while (start1--) fgets(buf, 1020, f1in);

  /* Read up to line start2 in file2 */
  while (start2--) fgets(buf, 1020, f2in);

  /* Now print out the lines */
  for (i = 0; i < maxlines; i++, num1--, num2--) {

    /* Print the left-hand line */
    if (num1 > 0) {
      fgets(buf, 1020, f1in);
      buf[strlen(buf) - 1] = '\0';	/* Remove newline */
      /* I used to printf("%-80s",buf); here, but it never did
       * what I wanted, so I now print out exactly 80 chars by hand. 
       * This is because we have to deal with tabs, dammit!
       */
      tab_upto=0; bptr=0;
      for (count_to_eighty=0;count_to_eighty<80;count_to_eighty++) {
        if (tab_upto>count_to_eighty) { fputc(' ', stdout); continue; }
	if (buf[bptr]=='\0') break;
	if (buf[bptr]=='\t') {
	  /* Calculate the next highest multiple of 8 */
	  tab_upto= (count_to_eighty+8) & ~7;
	  if (tab_upto>=80) break;
	  fputc(' ', stdout); bptr++; continue;
	}
	fputc(buf[bptr++], stdout);
      }
      for (;count_to_eighty<80;count_to_eighty++) {
	fputc(' ', stdout);
      }
      printf(" ");
    } else printf("%81s", " ");

    /* Print the right-hand line */
    if (num2 > 0) {
      fgets(buf, 1020, f2in);
      fputs(buf, stdout);
    } else printf("\n");
  }
  printf("\n");
  if (num1 > num2) printf("\n");
  fclose(f1in);
  fclose(f2in);
}


#ifdef ISOMORPH_CODE
/*
 * Here are some comments on the isomorph[] arrays. We would like to identify
 * code fragments that are similar, even if variables have been renamed or
 * even declared in a different order. For example, the following two
 * functions are essentially identical.
 * 
 * int maxofthree(int x, int y, int z)    int bigtriple(int b, int a, int c)
 * {                                      {
 *   if ((x>y) && (x>z)) return(x);          if ((a>b) && (a>c)) return(a);
 *   if (y>z) return(y);                     if (b>c) return(b);
 *   return(z);                              return(c);
 * }                                      }
 *
 * The tokenised forms of these functions are obviously different
 *
 * int id0(int id1,int id2,int id3)         int id0(int id1,int id2,int id3)
 * {                                        {
 *  if ((id1>id2)&&(id1>id3))                if ((id2>id1)&&(id2>id3))
 *    return(id1);                             return(id2);
 *  if(id2>id3) return(id2);                 if (id1>id3) return(id1);
 *  return(id3);                             return(id3);
 * }                                        }
 *
 * 
 * However, it is possible to compare these and show that they are identical. If
 * we start with the first "if" statement, a straight lexical comparison
 * shows that they are different. But at the first difference (id1 cf. id2),
 * we put an entry into the isodtos[] array so that isodtos[2]= 1, i.e. the
 * identifier #2 on the RHS is isomorphic to identifier #1 on the LHS. Thus
 * when we get to the first "return" statements, we can see that they are
 * also isomorphic. We can thus conclude that everything from the "if" down
 * to the "}" is isomorphic and thus identical.
 * 
 */

/*
 * isomorphic_compare(): compare two runs of identical tokens to see if they
 * are also isomorphic w.r.t the token values. Returns the number of
 * isomorphic tokens in the run starting at s and d.
 */
int
isomorphic_compare(struct token * s, struct token * d, int runlength)
{
  /* We need two tables: one to record the match from a dst value */
  /* to a src value, and another to record the src to dst value match */
  unsigned short isodtos[65536];
  unsigned short isostod[65536];
  int runcount;

  /* Clear the identifer isomorph table for this run */
  memset(isodtos, 0, sizeof(isodtos));
  memset(isostod, 0, sizeof(isostod));

  /* Walk the token runs from start to end */
  for (runcount=0; runcount<runlength; runcount++) {
    if (s->type != d->type) {
      printf("Assert error! token mismatch %d vs %d!!\n", s->type, d->type);
    }
    switch (s->type) {
    case IDENTIFIER:

      /* First thing, record a mapping each way if there is none */
      if (isodtos[d->value] == 0) isodtos[d->value] = s->value;
      if (isostod[s->value] == 0) isostod[s->value] = d->value;

      /* Now reject if the mappings fail in either direction */
      if (isodtos[d->value] != s->value) return(runcount);
      if (isostod[s->value] != d->value) return(runcount);
      break;
    case INTVAL:
      if (d->value != s->value) return (runcount);
      break;
    }
    s=s->next; d=d->next;
  }
  return (runcount);
}
#endif


/*
 * Compare the two source trees. We walk the linked token list we are given,
 * and then call another function one token at a time to find matches.
 */

void
compare_trees(void)
{
  unsigned int srcstart, dststart;
  unsigned int i, s, d;
  int runcount, isoruncount;
  int lastline;
  struct token *T;
#ifdef OPT_OVERLAP
  struct tokrun *Runhead=NULL;
  struct tokrun *Runtail=NULL;
  struct tokrun *prev, *this;
  int found;
#endif

  /* Walk the source int array */
  for (srcstart = 0; srcstart < Toksrcsize; srcstart++) {

    if ((progress==1) && (Toksrcref[srcstart]->type==STARTFILE1)) {
      fprintf(stderr, "Processing src file %s\n", Toksrcref[srcstart]->file);
    }

#ifdef OPT_OVERLAP
    /* Clean out the old tokrun list if this is a new srcfile */
    if (Toksrcref[srcstart]->type==STARTFILE1) {
      for (this=Runhead;this!=NULL;this=prev) {
	/* printf("Removing tokrun node 0x%x\n", (unsigned int)this); */
	prev=this->next; free(this);
      }
      Runhead=Runtail=NULL;
    }
#endif

    for (dststart = 0; dststart < Tokdstsize; dststart++) {

      s = srcstart; d = dststart; runcount = 0;

      /* We find and compare in steps of 4 tokens, which */
      /* suits 32-bit integer compares. */
      while (1) {
	if (Toksrcarray[s] != Tokdstarray[d]) break;
	runcount += 4; s += 4; d += 4;
      }

      /* Ignore the run if it is smaller than the threshold */
      /* We can't move dststart up to d, because we may miss */
      /* a different run where the srcstart token matches a */
      /* token between dststart and d which is itself a long run. */
      if (runcount < threshold) continue;

      /* Back up a bit and do the same one token at a time */
      s -= 3; d -= 3;
      while (1) {
	if (Toksrcarray[s] != Tokdstarray[d]) break;
	runcount++; s++; d++;
      }

#ifdef OPT_OVERLAP
      /* First check and ignore this if it's already recorded */
      found=0;
      for (this=Runhead; this!=NULL; this=this->next) {
	if ((srcstart>=this->sstart) && (s<=this->send) &&
	    (dststart>=this->dstart) && (d<=this->dend)) {
	  /* printf("Found an existing tokrun!\n"); */
	  found=1; break;
	}
      }
      if (found==1) continue;
     
      /* We record the run in the tokrun list */
      this= (struct tokrun *)malloc(sizeof(struct tokrun));
      if (this!=NULL) {
	this->sstart= srcstart; this->send=s;
	this->dstart= dststart; this->dend=d;
	this->next=NULL;
  
        /* Link into the list */
	/* printf("Adding tokrun node %d-%d  %d-%d\n",srcstart,s,dststart,d); */
	
	if (Runhead==NULL) { Runhead=Runtail=this; }
	else { Runtail->next= this; Runtail=this; }
      }
#endif

      /* At this point we have a possible token run of at least */
      /* threshold in length. We now need to apply some filtering */
      /* to weed out lots of false positives. */

#ifdef ISOMORPH_CODE
      isoruncount=isomorphic_compare(Toksrcref[srcstart],
				Tokdstref[dststart], runcount);
      if (isoruncount < threshold) continue;

      /* isomorphic_compare() might have found a run shorter than that */
      /* which we found in the beginning. Modify runcount to reflect this. */
      runcount= isoruncount;
#endif
      printf("%d  %s:%d  %s:%d\n", runcount,
	     Toksrcref[srcstart]->file,
	     Toksrcref[srcstart]->linenum,
	     Tokdstref[dststart]->file,
	     Tokdstref[dststart]->linenum);

      if (showtokens) {
	/* Now print out the list of tokens */
	lastline= Toksrcref[srcstart]->linenum;
	for (s = srcstart, i = 0; i < runcount; i++, s++) {
	  T = Toksrcref[s];
	  if (Toksrcref[s]->linenum > lastline) {
	    lastline= Toksrcref[s]->linenum;
	    fputc('\n', stdout);
	  }
	  fputs(tokstring[T->type], stdout);
	  switch (T->type) {
	  case IDENTIFIER:
	  case INTVAL:
	    printf("%d ", T->value);
	    break;
	  }
	}
	fputc('\n', stdout);
	fputc('\n', stdout);
      }
      if (crosscompare) {
	paste_files(Toksrcref[srcstart]->file, Toksrcref[srcstart]->linenum,
		    Toksrcref[s]->linenum - Toksrcref[srcstart]->linenum + 1,
		    Tokdstref[dststart]->file, Tokdstref[dststart]->linenum,
		    Tokdstref[d]->linenum - Tokdstref[dststart]->linenum + 1);
      }
    }
  }
}

void
usage(void)
{
  fprintf(stderr,
       "Usage: ctcompare [-p] [-t] [-x] [-i] [-n nnn] cft_file ctf_file\n");
  exit(1);
}


int
main(int argc, char *argv[])
{
  struct token *Srctoklist, *Dsttoklist;
  int ch;

  /* Process options */
  while ((ch = getopt(argc, argv, "ptxin:")) != -1) {
    switch (ch) {
    case 'p':
      progress = 1; break;
    case 'i':
      skip_cppstuff = 0; break;
    case 't':
      if (crosscompare) {
	fprintf(stderr, "-t and -x are mutually exclusive\n"); crosscompare = 0;
      }
      showtokens = 1; break;
    case 'x':
      if (showtokens) {
	fprintf(stderr, "-t and -x are mutually exclusive\n"); showtokens = 0;
      }
      crosscompare = 1; break;
    case 'n':
      threshold = atoi(optarg);
      if (threshold <= 0) {
	fprintf(stderr, "Bad value for -n\n"); threshold = 10;
      }
      break;
    default:
      usage();
    }
  }
  argc -= optind; argv += optind;

  if (argc != 2) usage();

  Srctoklist = load_ctffile(argv[0], 0);
  if (Srctoklist == NULL) {
    fprintf(stderr, "Error reading ctf file %s\n", argv[0]); exit(1);
  }
  Dsttoklist = load_ctffile(argv[1], 1);
  if (Dsttoklist == NULL) {
    fprintf(stderr, "Error reading ctf file %s\n", argv[1]); exit(1);
  }
  compare_trees();
  exit(0);
}
